import json
import os
import unittest

from worker.data.activity import Activity
from worker.data.activity.activity_grouping import ActivityGrouping, ActivityGroup
from worker.data.serialization import obj2json, json2obj
from worker.data.wits import WITS


class TestActivity(unittest.TestCase):
    # Mapping of merging activities:
    # key: merging group name
    # value: list of constituent activities that can be merged
    ACTIVITY_MERGING_DICTIONARY = {
        Activity.RUN_IN_HOLE: [Activity.RUN_IN_HOLE, Activity.DRY_REAMING_DOWN],
        Activity.PULL_OUT_OF_HOLE: [
            Activity.PULL_OUT_OF_HOLE, Activity.WASHING_UP,
            Activity.DRY_REAMING_UP, Activity.REAMING_UP
        ],
        Activity.CIRCULATING: [
            Activity.ROTARY_DRILLING, Activity.SLIDE_DRILLING,
            Activity.WASHING_DOWN, Activity.REAMING_DOWN,
            Activity.STATIC_OFF_BOTTOM, Activity.CIRCULATING,
            Activity.DRY_ROTARY_OFF_BOTTOM, Activity.ROTARY_OFF_BOTTOM,
            Activity.STATIC_ON_BOTTOM, Activity.CIRCULATING_ON_BOTTOM,
            Activity.DRY_ROTARY_ON_BOTTOM, Activity.CIRCULATING_AND_ROTARY_ON_BOTTOM,
            Activity.IN_SLIPS, Activity.UNCLASSIFIED
        ]
    }

    def test_activity_setting(self):
        activity = Activity("Reaming Down")
        self.assertEqual(Activity.REAMING_DOWN, activity)
        self.assertTrue(activity.is_reaming())
        self.assertTrue(activity.is_pipe_moving())
        self.assertTrue(activity.is_pumping())
        self.assertTrue(activity.is_rotating())

        activity = Activity("Rotary Off Bottom")
        self.assertNotEqual(Activity.REAMING_DOWN, activity)
        self.assertEqual(Activity.ROTARY_OFF_BOTTOM, activity)
        self.assertEqual("Rotary Off Bottom", activity.value)
        self.assertNotEqual("Rotary Drilling", activity.value)
        self.assertFalse(activity.is_reaming())
        self.assertFalse(activity.is_pipe_moving())
        self.assertTrue(activity.is_pumping())
        self.assertTrue(activity.is_rotating())

        activity = Activity("Rotary Drilling")
        self.assertNotEqual(Activity.WASHING_DOWN, activity)
        self.assertEqual(Activity.ROTARY_DRILLING, activity)
        self.assertEqual("Rotary Drilling", activity.value)
        self.assertFalse(activity.is_reaming())
        self.assertTrue(activity.is_pipe_moving())
        self.assertTrue(activity.is_pumping())
        self.assertTrue(activity.is_rotating())

    def test_activity_grouping(self):
        wits_records = get_data_for_activity_grouping()
        self.assertEqual(1714, len(wits_records))
        activity_grouping = ActivityGrouping(min_group_duration=5)
        groups = activity_grouping.group(wits_records)
        print()
        print("\n".join(str(grp) for grp in groups))
        self.assertEqual(18, len(groups))
        # the index of the groups and their expected results
        expected = {
            0: ActivityGroup(Activity('In Slips'), 1594037326, 1594037418),
            1: ActivityGroup(Activity('Static Off Bottom'), 1594037422, 1594037443),
            16: ActivityGroup(Activity('Rotary Off Bottom'), 1594038755, 1594038891),
            17: ActivityGroup(Activity('Rotary Drilling'), 1594038893, 1594039039)
        }
        for index, activity_group in expected.items():
            self.assertEqual(groups[index], activity_group)

    def test_activity_grouping_with_merging(self):
        wits_records = get_data_for_activity_grouping()

        self.assertEqual(1714, len(wits_records))

        activity_grouping = ActivityGrouping(min_group_duration=5, merging_dict=self.ACTIVITY_MERGING_DICTIONARY)
        groups = activity_grouping.group(wits_records)
        print()
        print("\n".join(str(grp) for grp in groups))
        # the index of the groups and their expected results
        expected = {
            0: ActivityGroup(Activity.CIRCULATING, 1594037326, 1594037418),
            1: ActivityGroup(Activity.CIRCULATING, 1594037422, 1594037443),
            2: ActivityGroup(Activity.RUN_IN_HOLE, 1594037444, 1594037541),
            3: ActivityGroup(Activity.CIRCULATING, 1594037542, 1594037646),
            4: ActivityGroup(Activity.RUN_IN_HOLE, 1594037650, 1594037800),
            5: ActivityGroup(Activity.CIRCULATING, 1594037801, 1594038017),
            6: ActivityGroup(Activity.CIRCULATING, 1594038022, 1594038582),
            7: ActivityGroup(Activity.CIRCULATING, 1594038586, 1594038751),
            8: ActivityGroup(Activity.CIRCULATING, 1594038755, 1594039039),
        }
        self.assertEqual(len(expected), len(groups))
        for index, activity_group in expected.items():
            self.assertEqual(groups[index], activity_group)

    def test_activity_serialization(self):
        washing_down = Activity.WASHING_DOWN
        ser = obj2json(washing_down)
        washing_down2 = json2obj(ser)
        self.assertEqual(washing_down, washing_down2)

    def test_merging_inslips(self):
        activities = ['In Slips'] + ['Run In Hole'] * 4 + ['In Slips'] + ['Run In Hole'] * 6 + ['In Slips'] * 18
        timestamps = list(range(len(activities)))

        wits_data = []
        for index in timestamps:
            wits_json = {'data': {'entry_at': timestamps[index], 'state': activities[index]}}
            w = WITS(wits_json)
            wits_data.append(w)
        activity_grouper = ActivityGrouping(
            min_group_duration=7,
            merging_dict=self.ACTIVITY_MERGING_DICTIONARY
        )
        activity_groups = activity_grouper.group(wits_data)
        self.assertEqual(1, len(activity_groups))
        self.assertEqual(18, activity_groups[0].duration)

    def test_activity_group_overlap(self):
        group1 = ActivityGroup(Activity.WASHING_DOWN, start=1, end=10)
        group2 = ActivityGroup(Activity.WASHING_DOWN, start=5, end=15)
        group3 = ActivityGroup(Activity.WASHING_DOWN, start=20, end=23)

        self.assertTrue(ActivityGroup.has_overlap(group1, group2))
        self.assertFalse(ActivityGroup.has_overlap(group2, group3))


def get_data_for_activity_grouping():
    wits_chunk = os.path.abspath(
        os.path.join(os.path.dirname(__file__), '..', 'resources', 'test', 'activity_grouping.data')
    )

    with open(wits_chunk, 'r') as file:
        lines = file.readlines()
        wits_records = [WITS(json.loads(r)) for r in lines]

    return wits_records



from typing import Union, List

from worker.data.operations import get_data_by_path, get_config_by_id, compare_float

from worker import API
from worker.exceptions import MissingConfigError

from worker.wellbore.model.drillstring import Drillstring
from worker.wellbore.model.enums import HoleType
from worker.wellbore.model.hole import Hole
from worker.wellbore.model.hole_section import HoleSection
from worker.wellbore.model.riser import Riser
from worker.wellbore.wellbore import Wellbore

MIN_LENGTH = 1  # ft


def run_drillstring_and_create_wellbore(wits: dict, mud_flow_in: float = None) -> Union[Wellbore, None]:
    """
    Get the wits record and create a wellbore
    :param wits: wits records in json format
    :param mud_flow_in: used to determine if under-reamer needs to be activated for hole enlargement
    :return: a wellbore
    """
    if wits is None:
        return None

    api_worker = API()

    drillstring_id = get_data_by_path(wits, "metadata.drillstring", str, default=None)

    if not drillstring_id:
        return None

    drillstring = get_config_by_id(drillstring_id, collection="data.drillstring")

    if not drillstring:
        return None

    active_drillstring_number = drillstring.get("data", {}).get("id")

    # setting active drillstring
    drillstring = Drillstring(drillstring)

    bit = drillstring.get_bit()
    if not bit:
        raise MissingConfigError(f"Bit was not found in the drillstring with _id='{drillstring_id}'")

    asset_id = wits['asset_id']
    bit_size = bit.size

    bit_depth = get_data_by_path(wits, "data.bit_depth", float)
    hole_depth = get_data_by_path(wits, "data.hole_depth", float)

    # setting cased-hole sections
    query = "{data.inner_diameter#gte#%s}" % bit_size
    casings = api_worker.get(path="/v1/data/corva", collection="data.casing", asset_id=asset_id, query=query, limit=100).data
    hole = set_casings(casings)
    smallest_cased_hole_diameter = hole.get_min_inner_diameter() or 100

    # ====== setting open-hole sections
    query = "{data.id#lte#%s}" % active_drillstring_number
    sort = "{data.id:1}"
    drillstrings = api_worker.get(path="/v1/data/corva", collection="data.drillstring", asset_id=asset_id, query=query, sort=sort, limit=100).data
    hole = add_open_holes(hole, drillstrings, smallest_cased_hole_diameter, hole_depth)

    mud_flow_in = get_mud_flow_in(mud_flow_in, wits)

    return Wellbore(
        drillstring=drillstring,
        hole=hole,
        bit_depth=bit_depth,
        hole_depth=hole_depth,
        mud_flow_in=mud_flow_in,
    )


def run_casingstring_and_create_wellbore(wits: dict, mud_flow_in: float = None) -> Union[Wellbore, None]:
    """
    Get the wits record and create a wellbore
    :param wits: wits records in json format
    :param mud_flow_in: used to determine if under-reamer needs to be activated for hole enlargement
    :return: a wellbore
    """
    if wits is None:
        return None

    api_worker = API()

    casingstring_id = get_data_by_path(wits, "metadata.casing", str, default=None)

    if not casingstring_id:
        return None

    casingstring = get_config_by_id(casingstring_id, collection="data.casing")

    if not casingstring:
        return None

    casing_outer_diameter = casingstring.get("data", {}).get("outer_diameter")
    casing_inner_diameter = casingstring.get("data", {}).get("inner_diameter")

    # setting active drillstring
    casingstring = Drillstring(casingstring)

    asset_id = wits['asset_id']

    bit_depth = get_data_by_path(wits, "data.bit_depth", float)
    hole_depth = get_data_by_path(wits, "data.hole_depth", float)

    # setting cased-hole sections
    query = "{data.inner_diameter#gt#%s}" % casing_inner_diameter
    sort = "{data.inner_diameter:-1}"
    casings = api_worker.get(path="/v1/data/corva", collection="data.casing", asset_id=asset_id, query=query, sort=sort, limit=100).data

    # ====== setting open-hole sections
    sort = "{timestamp: -1}"
    drillstrings = api_worker.get(path="/v1/data/corva", collection="data.drillstring", asset_id=asset_id, sort=sort, limit=100).data

    drillstrings = [
        next((string for string in drillstrings if (string.get("data", {}).get("components", [{}])[-1].get("size") or -1) > casing_outer_diameter), {})
    ]

    hole = set_casings(casings)
    smallest_cased_hole_diameter = hole.get_min_inner_diameter() or 100

    hole = add_open_holes(hole, drillstrings, smallest_cased_hole_diameter, hole_depth)

    mud_flow_in = get_mud_flow_in(mud_flow_in, wits)

    return Wellbore(
        drillstring=casingstring,
        hole=hole,
        bit_depth=bit_depth,
        hole_depth=hole_depth,
        mud_flow_in=mud_flow_in,
    )


def get_detailed_hole_from_elements(elements: List[dict],
                                    previous_element_top_depth: float,
                                    possible_top_depth: float,
                                    extend_casing_to_top: bool,
                                    is_riser: bool
                                    ) -> Hole:
    """
    This function converts elements of casing/riser components into a Hole.
    :param elements: list casing/riser components
    :param previous_element_top_depth: will be used as the bottom depth of current component
    :param possible_top_depth:  possible top depth to which a casing/riser can extend. In case of a casing, it will extend to the btm of riser (if available)
    :param extend_casing_to_top: if the casing/riser should extend to the top
    :param is_riser: If its a riser or a cased hole
    :return: Hole with all the components as HoleSections
    """
    casing_hole: Hole = Hole()

    elements = [element for index, element in enumerate(elements) if check_casing_element_validity(element, index, extend_casing_to_top)]

    for element in reversed(elements):
        element_id = element.get("inner_diameter", 0.0)
        element_length = element.get("length", 0.0)
        element_bottom = previous_element_top_depth
        element_top = element_bottom - element_length

        if element_top < possible_top_depth:
            create_section_and_append(casing_hole, element_id, possible_top_depth, element_bottom, HoleType.CASED_HOLE, is_riser)
            break

        create_section_and_append(casing_hole, element_id, element_top, element_bottom, HoleType.CASED_HOLE, is_riser)
        previous_element_top_depth = element_top

    casing_hole.sections.reverse()

    # checking and making sure the top element extends to the possible top depth.
    # This check is needed because we are ignoring components of dp family which maybe at the top of the liner.
    if compare_float(casing_hole.sections[0].top_depth, possible_top_depth, tolerance=0.01) != 0:
        casing_hole.sections[0].set_top_depth(possible_top_depth)
        casing_hole.sections[0].set_length()

    return casing_hole


def create_section_and_append(casing_hole, inner_diameter, top_depth, bottom_depth, hole_type, is_riser):
    sec = HoleSection(inner_diameter=inner_diameter, top_depth=top_depth, bottom_depth=bottom_depth, hole_type=hole_type)
    if is_riser:
        sec = Riser(inner_diameter=inner_diameter, top_depth=top_depth, bottom_depth=bottom_depth, hole_type=hole_type)

    casing_hole.sections.append(sec)


def check_casing_element_validity(casing_element: dict, idx: int, extend_casing_to_top: bool) -> bool:
    """
    Check if the component is a casing and of a significant length
    :param casing_element:
    :param idx:
    :param extend_casing_to_top:
    :return:
    """
    element_length: float = casing_element.get("length")
    element_family: str = casing_element.get("family")

    if not element_family or not isinstance(element_family, str):
        return False

    ignored_families = (
        "dp",
        "hwdp"
    )

    # A dp component in casing is considered invalid. It is ignored when setting casing
    if element_family.lower() in ignored_families:
        return False

    # Ignoring casing components with length less than 20 feet and only components that are not top most or those that don't extend to top depth.
    if element_length < 20.0 and (idx != 0 or not extend_casing_to_top):
        return False

    return True


def set_casings(casings: List[dict]) -> Hole:
    """
    Set the casing by providing the list of the jsons
    :param casings: list of casings from API in json dict format
    :return: None
    """
    if not casings:
        return Hole()

    possible_casing_top_depth: float = 0.0
    riser = next((csg for csg in casings if csg.get('data', {}).get('is_riser', False)), {})

    if riser:
        possible_casing_top_depth = riser.get("data").get("bottom_depth")
        casings.remove(riser)

    # sorting the casings in the ascending order (the smaller the inner_diameter the earlier it is added)
    casings = sorted(casings, key=lambda csg: csg.get('data', {}).get('inner_diameter'))

    # finding if a liner is overlapping outer liners
    # this is common in offshore drilling where they set the casings at the sea floor
    index = 1
    while index < len(casings):
        top_depth_inner = casings[index - 1].get('data', {}).get('top_depth', 1000)
        top_depth_outer = casings[index].get('data', {}).get('top_depth', 1000)
        if top_depth_outer + Wellbore.DEPTH_THRESHOLD > top_depth_inner:
            casings.pop(index)
        else:
            index += 1

    # reversing the order (the larger the inner_diameter the earlier it is)
    casings = reversed(casings)

    casing_holes = []
    for index, casing in enumerate(casings):
        casing_hole = Hole()
        casing_object: dict = casing.get("data", {})
        casing_bottom = casing_object.get("bottom_depth")
        casing_top = possible_casing_top_depth
        current_casing_top_depth = casing_object.get("top_depth")

        # The top depth of the first component should reach to surface regardless
        # Case 1 - Land (Normal)
        # Case 2 - Riser
        # Case 3 - Riserless
        if not (index == 0 and current_casing_top_depth < Wellbore.DEPTH_THRESHOLD):
            casing_top = current_casing_top_depth

        casing_elements: List[dict] = casing_object.get("components", [])
        if not casing_elements:
            hole_section = HoleSection(**casing_object, hole_type=HoleType.CASED_HOLE)
            casing_hole.sections.append(hole_section)
        else:
            previous_casing_element_top_depth = casing_bottom
            # Individual casing elements have to be looped in reverse order to correctly account for the topmost element going to surface is not liner
            extend_casing_to_top: bool = False
            if index == 0:
                extend_casing_to_top = True
            casing_hole = get_detailed_hole_from_elements(casing_elements, previous_casing_element_top_depth, casing_top, extend_casing_to_top, False)

        casing_holes.append(casing_hole)

    # Now adding the components of the riser to casingHoles
    if riser:
        riser_elements: List[dict] = riser.get("data", {}).get("components", [])
        previous_riser_element_top_depth = possible_casing_top_depth
        riser_hole: Hole = get_detailed_hole_from_elements(riser_elements, previous_riser_element_top_depth, 0.0, True, True)
        casing_holes.insert(0, riser_hole)

    return Hole.merge_holes(casing_holes)


def add_open_holes(hole: Hole, drillstrings: List[dict], smallest_cased_hole_diameter: float, hole_depth: float) -> Hole:
    for ds in drillstrings:
        bit = Drillstring(ds).get_bit()
        if not bit:
            drillstring_id = ds.get("_id")
            raise MissingConfigError(f"Bit was not found in the drillstring with _id='{drillstring_id}'")

        bit_size = bit.size
        # only keeping the DSs that ran below the smallest casing
        if bit_size > smallest_cased_hole_diameter:
            continue

        start_depth = ds.get("data", {}).get("start_depth")

        # if due to the side track the hole depth reduced, then it should update the setting depths in the open hole
        hole.sections = [
            sec for sec in hole.sections
            if sec.hole_type != HoleType.OPEN_HOLE or sec.top_depth <= hole_depth
        ]

        # if a new open-hole section is created in addition to the previous one,
        # the previous bottom depth may need to be modified.
        previous_hole_section = hole.get_last_hole_section()

        top_depth = hole.get_bottom_depth()
        bottom_depth = top_depth + 0.0001
        current_hole_section = HoleSection(
            top_depth=top_depth,
            bottom_depth=bottom_depth,
            inner_diameter=bit_size,
            hole_type=HoleType.OPEN_HOLE
        )

        if previous_hole_section:
            # The casing bottom depth can't be modified if the difference is more than 300 ft
            cased_hole_condition = previous_hole_section.hole_type == HoleType.CASED_HOLE and start_depth - previous_hole_section.bottom_depth < 300
            if cased_hole_condition or previous_hole_section.hole_type == HoleType.OPEN_HOLE:
                previous_hole_section.set_bottom_depth(start_depth)
                previous_hole_section.set_length()

            if current_hole_section.eq_without_length(previous_hole_section):
                continue

        current_hole_section.set_length()
        hole.add_section(current_hole_section)

    # In some cases the DS setting depth doesn't match the hole depth so the wits data overrides it;
    # this is only applied to the top depth of the last section and bottom depth of the previous section.
    # Note: only in cases that the wits hole_depth < setting depth
    # Condition: the number of open hole sections should be at least two needs to be satisfied first
    if len(hole) - len(hole.get_cased_hole()) > 1 and 1 < hole_depth < hole[-1].top_depth:
        hole[-2].set_bottom_depth(hole_depth)
        hole[-2].set_length()

        hole[-1].set_top_depth(hole_depth)
        hole[-1].set_length()

    return hole


def get_mud_flow_in(mud_flow_in: Union[float, None], wits: dict) -> Union[float, None]:
    """
    Get the mud flow rate from wits if `mud_flow_in` is None
    :param mud_flow_in:
    :param wits:
    :return:
    """
    if mud_flow_in is not None:
        return mud_flow_in

    if not wits:
        return None

    return get_data_by_path(wits, "data.mud_flow_in", default=None)



import json
import os
import unittest

from worker import API
from worker.data.alert import Alerter
from worker.data.operations import get_data_by_path, get_one_data_record, get_config_by_id
from worker.data.serialization import obj2json, json2obj
from worker.mixins.logging import Logger
from worker.test.utils import file_to_json
from worker.wellbore.factory import run_drillstring_and_create_wellbore, run_casingstring_and_create_wellbore
from worker.wellbore.measured_depth_finder import get_unique_measured_depths
from worker.wellbore.model.annulus import Annulus
from worker.wellbore.model.drillstring import Drillstring
from worker.wellbore.model.enums import HoleType, PipeType
from worker.wellbore.model.hole import Hole
from worker.wellbore.model.hole_section import HoleSection


class TestSections(unittest.TestCase):
    def test_hole_section(self):
        hole_section1 = HoleSection(inner_diameter=5, top_depth=0, bottom_depth=20_000, hole_type=HoleType.OPEN_HOLE)

        hole_section1.set_length()

        self.assertEqual(20_000, hole_section1.length)

        expected_area = 19.634954084936208  # in INCH^2
        expected_volume = 2727.07695624114  # in FOOT^3

        self.assertAlmostEqual(expected_area, hole_section1.compute_get_area(), places=10)
        self.assertAlmostEqual(expected_volume, hole_section1.compute_get_volume(), places=10)

    def test_hole(self):
        casings = os.path.abspath(
            os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources", "test", "casings.json")
        )
        casings = file_to_json(casings)

        hole = Hole()
        hole.set_casings(casings)
        self.assertEqual(2, len(hole))

        hole_section1 = hole[0]
        self.assertAlmostEqual(6.88, hole_section1.inner_diameter, 10)
        self.assertAlmostEqual(6_770, hole_section1.length, 10)
        self.assertAlmostEqual(0, hole_section1.top_depth, 10)
        self.assertAlmostEqual(6_770, hole_section1.bottom_depth, 10)
        self.assertEqual(HoleType.CASED_HOLE, hole_section1.hole_type)

        hole_section2 = hole[1]
        self.assertAlmostEqual(4.67, hole_section2.inner_diameter, 10)
        self.assertAlmostEqual(14192, hole_section2.length, 10)
        self.assertAlmostEqual(6_770, hole_section2.top_depth, 10)
        self.assertAlmostEqual(20_962, hole_section2.bottom_depth, 10)
        self.assertEqual(HoleType.CASED_HOLE, hole_section2.hole_type)

        self.assertAlmostEqual(20_962, hole.get_bottom_depth(), 10)

        # Open hole
        open_hole = HoleSection(inner_diameter=4.5, top_depth=1111, bottom_depth=21_000, hole_type=HoleType.OPEN_HOLE)
        hole.add_section(open_hole)
        hole_section3 = hole[2]
        self.assertAlmostEqual(4.5, hole_section3.inner_diameter, 10)
        self.assertAlmostEqual(21_000 - 20_962, hole_section3.length, 10)
        self.assertAlmostEqual(20_962, hole_section3.top_depth, 10)
        self.assertAlmostEqual(21_000, hole_section3.bottom_depth, 10)
        self.assertEqual(HoleType.OPEN_HOLE, hole_section3.hole_type)

        self.assertAlmostEqual(21_000, hole.get_bottom_depth(), 10)

        # test fetching of the correct casing at a given depth
        casing1 = hole.find_section_at_measured_depth(10_000, False)
        self.assertEqual(id(hole_section2), id(casing1))

        # testing unique boundary depths
        mds = get_unique_measured_depths(hole)
        self.assertListEqual(mds, [0, 6770.0, 20962.0, 21000.0])

        # test serialization
        ser_str = obj2json(hole)
        hole_des = json2obj(ser_str)
        self.assertEqual(len(hole), len(hole_des))
        self.assertAlmostEqual(hole_des[1].inner_diameter, hole[1].inner_diameter, 10)
        self.assertAlmostEqual(hole_des[1].top_depth, hole[1].top_depth, 10)

        print("Hole:")
        print(hole)
        return hole

    def test_drillstring(self):
        drillstring_json_file = os.path.abspath(
            os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources", "test", "drillstring.json")
        )
        drillstring_json = file_to_json(drillstring_json_file)

        drillstring = Drillstring()
        drillstring.set_drillstring(drillstring_json)

        component_1 = drillstring[0]
        self.assertAlmostEqual(3.826, component_1.inner_diameter, 10)
        self.assertAlmostEqual(4.5, component_1.outer_diameter, 10)
        self.assertAlmostEqual(31, component_1.joint_length, 10)
        self.assertAlmostEqual(31, component_1.length, 10)

        component_6 = drillstring[5]
        self.assertAlmostEqual(2.75, component_6.inner_diameter, 10)
        self.assertAlmostEqual(4.75, component_6.outer_diameter, 10)
        self.assertAlmostEqual(29.09, component_6.joint_length, 10)
        self.assertAlmostEqual(29.09, component_6.length, 10)

        bit = drillstring[-1]
        self.assertAlmostEqual(6.75, bit.size, 10)
        self.assertAlmostEqual(1.1780972450961724, bit.tfa, 10)
        self.assertEqual(PipeType.BIT.value, bit.pipe_type.value, 10)

        # test for adjusting depths using bit depth
        wits = get_wits()
        drillstring.update(wits)
        self.assertAlmostEqual(wits.get('data', {}).get('bit_depth'), drillstring[-1].bottom_depth, 10)

        # test for finding drillstring section at a measured depth
        drillstring_section3 = drillstring.find_section_at_measured_depth(8_000, False)
        self.assertAlmostEqual(3.83, drillstring_section3.inner_diameter, 1)
        self.assertAlmostEqual(4.5, drillstring_section3.outer_diameter, 10)

        print("Drillstring")
        print(drillstring)
        return drillstring

    def test_annulus(self):
        drillstring = self.test_drillstring()
        hole = self.test_hole()
        annulus = Annulus(drillstring=drillstring, hole=hole)
        print("Annulus:")
        print(annulus)

        # 2nd DP should be split into two segments
        section2, section3 = annulus[2: 4]
        self.assertAlmostEqual(6.88, section2.inner_diameter_hole, 10)
        self.assertEqual(HoleType.CASED_HOLE, section2.hole_type)
        self.assertAlmostEqual(4.67, section3.inner_diameter_hole, 10)
        self.assertEqual(HoleType.CASED_HOLE, section3.hole_type)

        # the segment above the is extended to the bit depth
        last_section = annulus[-1]

        component_above_bit, bit = drillstring[-2:]
        combined_length = component_above_bit.length + bit.length

        self.assertAlmostEqual(combined_length, last_section.length, 10)

    def test_wellbore_creation_with_factory(self):
        """
        In this test case the creation of the wellbore through
        factory is shown; note that only the wits record is used
        and the other elements (casings, drillstrings) are fetched
        from the API.
        :return:
        """
        wits = get_wits()
        bit_depth = get_data_by_path(wits, 'data.bit_depth', float)
        hole_depth = get_data_by_path(wits, 'data.hole_depth', float)
        wellbore = run_drillstring_and_create_wellbore(wits)

        print("\n Hole configuration is:")
        hole = wellbore.actual_hole
        print(hole)
        self.assertEqual(2, len(hole))
        self.assertEqual(HoleType.CASED_HOLE, hole[0].hole_type)
        self.assertEqual(HoleType.OPEN_HOLE, hole[1].hole_type)
        self.assertAlmostEqual(hole_depth, hole.get_bottom_depth(), 10)

        print("\n Drillstring configuration is:")
        drillstring = wellbore.actual_drillstring
        self.assertAlmostEqual(bit_depth, drillstring.bit_depth, 10)
        print(drillstring)

        print("\n Annulus configuration is:")
        annulus = wellbore.actual_annulus
        print(annulus)

        # testing drilling body volume change
        vol_change_calculated = wellbore.compute_get_drillstring_body_volume_change(6000, 5900)
        area_dp = wellbore.actual_drillstring[0].compute_body_cross_sectional_area_tj_adjusted()  # in INCH^2
        vol_expected = area_dp / 144 * -100
        self.assertAlmostEqual(vol_change_calculated, vol_expected, 10)

        # testing drilling outer solid volume change
        vol_change_calculated = wellbore.compute_get_drillstring_outside_volume_change(6000, 5900)
        area_dp = wellbore.actual_drillstring[0].compute_outer_area_tool_joint_adjusted()  # in INCH^2
        vol_expected = area_dp / 144 * -100
        self.assertAlmostEqual(vol_change_calculated, vol_expected, 10)

        # test serialization
        ser_str = obj2json(wellbore)
        wellbore_des = json2obj(ser_str)
        self.assertEqual(len(wellbore.actual_drillstring), len(wellbore_des.actual_drillstring))
        self.assertAlmostEqual(wellbore.actual_drillstring[1].inner_diameter, wellbore_des.actual_drillstring[1].inner_diameter, 10)
        self.assertAlmostEqual(wellbore.actual_drillstring[1].top_depth, wellbore_des.actual_drillstring[1].top_depth, 10)

    def test_wellbore_creation_with_under_reamer(self):
        """
        In this test case the creation of the wellbore through
        factory is shown; note that only the wits record is used
        and the other elements (casings, drillstrings) are fetched
        from the API.
        :return:
        """
        os.environ["LOGGING_LEVEL"] = "debug"
        os.environ["LOGGING_ASSET_ID"] = "all"

        asset_id = 81592982
        Logger.set_asset_id(asset_id)
        Alerter.set_asset_id(asset_id)

        case_results = {
            "flow_activated": {
                "hole": (
                    (HoleType.CASED_HOLE, 8.835, 0.0, 9890.0),
                    (HoleType.OPEN_HOLE, 8.75, 9890.0, 9891.0),
                    (HoleType.OPEN_HOLE, 8.5, 9891.0, 14901.07),
                    (HoleType.OPEN_HOLE, 12.0, 14901.07, 20539.77),
                    (HoleType.OPEN_HOLE, 8.5, 20539.77, 20633.0)
                ),
                "annulus": (
                    (HoleType.CASED_HOLE, 8.835, 5.0, 0, 9890.0),
                    (HoleType.OPEN_HOLE, 8.75, 5.0, 9890.0, 9891.0),
                    (HoleType.OPEN_HOLE, 8.5, 5.0, 9891.0, 14901.07),
                    (HoleType.OPEN_HOLE, 12.0, 5.0, 14901.07, 20528.9),
                    (HoleType.OPEN_HOLE, 12.0, 8.25, 20528.9, 20539.77),
                    (HoleType.OPEN_HOLE, 8.5, 6.75, 20539.77, 20633.0),
                )
            },
            "open_hole_activated": {
                "hole": (
                    (HoleType.CASED_HOLE, 8.835, 0.0, 9890.0),
                    (HoleType.OPEN_HOLE, 8.75, 9890.0, 9891.0),
                    (HoleType.OPEN_HOLE, 12.0, 9891.0, 20539.77),
                    (HoleType.OPEN_HOLE, 8.5, 20539.77, 20633.0)
                ),
                "annulus": (
                    (HoleType.CASED_HOLE, 8.835, 5.0, 0, 9890.0),
                    (HoleType.OPEN_HOLE, 8.75, 5.0, 9890.0, 9891.0),
                    (HoleType.OPEN_HOLE, 12.0, 5.0, 9891.0, 20528.9),
                    (HoleType.OPEN_HOLE, 12.0, 8.25, 20528.9, 20539.77),
                    (HoleType.OPEN_HOLE, 8.5, 6.75, 20539.77, 20633.0),
                )
            },
            "unactivated": {
                "hole": (
                    (HoleType.CASED_HOLE, 8.835, 0.0, 9890.0),
                    (HoleType.OPEN_HOLE, 8.75, 9890.0, 9891.0),
                    (HoleType.OPEN_HOLE, 8.5, 9891.0, 20633.0)
                ),
                "annulus": (
                    (HoleType.CASED_HOLE, 8.835, 5.0, 0, 9890.0),
                    (HoleType.OPEN_HOLE, 8.75, 5.0, 9890.0, 9891.0),
                    (HoleType.OPEN_HOLE, 8.5, 5.0, 9891.0, 20528.9),
                    (HoleType.OPEN_HOLE, 8.5, 8.25, 20528.9, 20539.77),
                    (HoleType.OPEN_HOLE, 8.5, 6.75, 20539.77, 20633.0),
                )
            },
            "depth_activation": {
                "hole": (
                    (HoleType.CASED_HOLE, 8.835, 0.0, 9890.0),
                    (HoleType.OPEN_HOLE, 8.75, 9890.0, 9891.0),
                    (HoleType.OPEN_HOLE, 8.5, 9891.0, 15205.17),
                    (HoleType.OPEN_HOLE, 12.0, 15205.17, 20539.77),
                    (HoleType.OPEN_HOLE, 8.5, 20539.77, 20633.0)
                ),
                "annulus": (
                    (HoleType.CASED_HOLE, 8.835, 5.0, 0, 9890.0),
                    (HoleType.OPEN_HOLE, 8.75, 5.0, 9890.0, 9891.0),
                    (HoleType.OPEN_HOLE, 8.5, 5.0, 9891.0, 15205.17),
                    (HoleType.OPEN_HOLE, 12.0, 5.0, 15205.17, 20528.9),
                    (HoleType.OPEN_HOLE, 12.0, 8.25, 20528.9, 20539.77),
                    (HoleType.OPEN_HOLE, 8.5, 6.75, 20539.77, 20633.0),
                )
            }
        }

        def get_under_reamer_from_api(mongo_id):
            drillstring = get_config_by_id(mongo_id, collection="data.drillstring")
            components = drillstring.get("data", {}).get("components", [])
            return next((component for component in components if component.get("family") == "ur"), {})

        for activation_type, expected_values in case_results.items():
            print(f"\nCase: Under-reamer is {activation_type}")
            # Set the environment variable to simulate that this is a task app
            os.environ["AWS_LAMBDA_FUNCTION_NAME"] = "unit-testing-lambda-task"

            wits_json_file = os.path.abspath(
                os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources", "test", "wits_under_reamer.json")
            )

            # Reset Drillstring
            reset_drillstring(activation_type)

            wits = file_to_json(wits_json_file)
            self.run_under_reamer_case(wits, expected_values)

            # Assert drillstring on API is not updated becuase it is not triggered from a stream app
            # The method above `reset_drillstring` resets the drillstring to original condition
            # The original condition for this test case are:
            #    UR_ACTIVE :   `ur_opened_depth` is set to `None`
            #    UR_INACTIVE:  `ur_opened_depth` is set to `10_000_000`
            # When running from this test case, the api should not be updated, and above values should be present on api
            # Following assertions ensure that. However when running from stream app, the api will be updated

            _id = wits["metadata"]["drillstring"]
            under_reamer_component = get_under_reamer_from_api(_id)

            if activation_type == "unactivated":
                self.assertEqual(10_000_000.0, under_reamer_component.get("ur_opened_depth"))
            elif activation_type == "depth_activation":
                self.assertEqual(15205.17, under_reamer_component.get("ur_opened_depth"))
            else:
                self.assertIsNone(under_reamer_component.get("ur_opened_depth"))

        print("\nCase: Running UR test to ensure that API is actually updated when running from a stream app")
        # Set the environment variable to simulate that this is not a task app
        os.environ["AWS_LAMBDA_FUNCTION_NAME"] = "unit-testing-lambda"
        expected_values = case_results.get("flow_activated")
        wits_json_file = os.path.abspath(
            os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources", "test", "wits_under_reamer.json")
        )
        reset_drillstring("flow_activated")
        wits = file_to_json(wits_json_file)
        self.run_under_reamer_case(wits, expected_values)
        _id = wits["metadata"]["drillstring"]
        under_reamer_component = get_under_reamer_from_api(_id)
        # Assert that api was actually updated in this test case becuase it simulates a stream app
        self.assertEqual(14901.07, under_reamer_component.get("ur_opened_depth"))

    def test_overlapping_liners(self):
        """
        In this test case we are checking if the casings and liners
        are set correctly even if the there are overlapping liners.
        :return:
        """

        wits = get_one_data_record(asset_id=3083, collection="wits")
        wellbore = run_drillstring_and_create_wellbore(wits)

        hole: Hole = wellbore.actual_hole
        hole_depth = 25633.01

        self.assertEqual(10, len(hole))

        self.assertEqual(HoleType.CASED_HOLE, hole[0].hole_type)
        self.assertAlmostEqual(0.0, hole[0].top_depth, 1)
        self.assertAlmostEqual(11863.3, hole[0].bottom_depth, 1)
        self.assertAlmostEqual(4.67, hole[0].inner_diameter, delta=0.01)

        self.assertEqual(HoleType.CASED_HOLE, hole[1].hole_type)
        self.assertAlmostEqual(11863.3, hole[1].top_depth, 1)
        self.assertAlmostEqual(11902.0, hole[1].bottom_depth, 1)
        self.assertAlmostEqual(4.28, hole[1].inner_diameter, delta=0.01)

        self.assertEqual(HoleType.CASED_HOLE, hole[8].hole_type)
        self.assertAlmostEqual(20610.0, hole[8].top_depth, 1)
        self.assertAlmostEqual(23000.0, hole[8].bottom_depth, 1)
        self.assertAlmostEqual(3.0, hole[8].inner_diameter, delta=0.01)

        self.assertEqual(HoleType.OPEN_HOLE, hole[9].hole_type)
        self.assertAlmostEqual(23000.0, hole[9].top_depth, 1)
        self.assertAlmostEqual(hole_depth, hole[9].bottom_depth, 1)
        self.assertAlmostEqual(2.85, hole[9].inner_diameter, delta=0.01)

    def test_creating_wellbore_with_missing_drillstring(self):
        wits_record = {
            "collection": "wits",
            "asset_id": 10477107,
            "metadata": {
                'drillstring': '5f4d5bf1150c40007f3a12fc'
            },
            "data": {
                "bit_depth": 1111,
                "hole_depth": 1111
            }
        }

        wellbore = run_drillstring_and_create_wellbore(wits_record)
        self.assertIsNone(wellbore)

    def test_multiple_casings(self):
        os.environ["LOGGING_LEVEL"] = "debug"
        os.environ["LOGGING_ASSET_ID"] = "all"

        # DS 5 of Odysseus well
        asset_id = 88739932
        Logger.set_asset_id(asset_id)

        wits = {
            "asset_id": asset_id,
            "metadata": {
                "drillstring": "5fda35ce1c5c89001ce626d2",
            },
            "data": {
                "bit_depth": 15_000.0,
                "hole_depth": 15_000.0
            }
        }

        expected_results = {
            "hole": (
                (HoleType.CASED_HOLE, 12.0, 0.0, 1500.0),  # Riser
                (HoleType.CASED_HOLE, 28.0, 1500.0, 1700.0),  # Conductor
                (HoleType.CASED_HOLE, 8.835, 1700.0, 9907.0),  # Intermediate casing
                (HoleType.OPEN_HOLE, 8.75, 9907.0, 15000.0),
            ),
            "annulus": (
                (HoleType.CASED_HOLE, 12, 5, 0.0, 1500.0),
                (HoleType.CASED_HOLE, 28, 5, 1500.0, 1700.0),
                (HoleType.CASED_HOLE, 8.835, 5, 1700.0, 9907.0),
                (HoleType.OPEN_HOLE, 8.75, 5, 9907.0, 14897.68),
                (HoleType.OPEN_HOLE, 8.75, 6.69, 14897.68, 14959.63),
                (HoleType.OPEN_HOLE, 8.75, 6.75, 14959.63, 14961.95),
                (HoleType.OPEN_HOLE, 8.75, 6.81, 14961.95, 14971.61),
                (HoleType.OPEN_HOLE, 8.75, 6.75, 14971.61, 15000.0)
            )
        }

        wellbore = run_drillstring_and_create_wellbore(wits)
        hole = wellbore.actual_hole
        annulus = wellbore.actual_annulus

        # Assert hole configuration
        print("\n Hole configuration is:")
        print(hole)
        self.assert_hole(expected_results.get("hole"), hole)

        # Assert annulus configuration
        print("\n Annulus configuration is:")
        print(annulus)
        self.assert_annulus(expected_results.get("annulus"), annulus)

    def test_consecutive_casings(self):
        os.environ["LOGGING_LEVEL"] = "debug"
        os.environ["LOGGING_ASSET_ID"] = "all"

        # Riser of Odysseus well
        asset_id = 88739932
        Logger.set_asset_id(asset_id)

        wits = {
            "asset_id": asset_id,
            "metadata": {
                "casing": "5fda3772a7e939001c289bd8"
            },
            "data": {
                "bit_depth": 1_000.0,
                "hole_depth": 2_338.0
            }
        }

        expected_results = {
            "hole": (
                (HoleType.CASED_HOLE, 28.0, 1500.0, 1700.0),  # Conductor casing
                (HoleType.CASED_HOLE, 12.415, 1700.0, 2338.0)  # Surface Casing
            ),
            "annulus": ()
        }

        wellbore = run_casingstring_and_create_wellbore(wits)
        hole = wellbore.actual_hole
        annulus = wellbore.actual_annulus

        # Assert hole configuration
        print("\n Hole configuration is:")
        print(hole)
        self.assert_hole(expected_results.get("hole"), hole)

        # Assert annulus configuration
        print("\n Annulus configuration is:")
        print(annulus)
        self.assert_annulus(expected_results.get("annulus"), annulus)

    def test_detailed_liner(self):
        os.environ["LOGGING_LEVEL"] = "debug"
        os.environ["LOGGING_ASSET_ID"] = "all"

        asset_id = 13684306
        Logger.set_asset_id(asset_id)

        wits = {
            "asset_id": asset_id,
            "metadata": {
                "drillstring": "5fbe0ec8eb6ace1f53efdd01"
            },
            "data": {
                "bit_depth": 9700.0,
                "hole_depth": 9700.0
            }
        }

        expected_results = {
            "hole": (
                (HoleType.CASED_HOLE, 8.84, 0.0, 5494.0),
                (HoleType.CASED_HOLE, 6.88, 5494.0, 8483.0),
                (HoleType.CASED_HOLE, 6.88, 8483.0, 8525.0),
                (HoleType.CASED_HOLE, 6.8, 8525.0, 8567.0),
                (HoleType.OPEN_HOLE, 6.75, 8567.0, 9700.0)
            ),
            "annulus": (
                (HoleType.CASED_HOLE, 8.84, 4.50, 0, 5494.0),
                (HoleType.CASED_HOLE, 6.88, 4.50, 5494.0, 8525.0),
                (HoleType.CASED_HOLE, 6.8, 4.50, 8525.0, 8567.0),
                (HoleType.OPEN_HOLE, 6.75, 4.50, 8567.0, 9581.61),
                (HoleType.OPEN_HOLE, 6.75, 5.25, 9581.61, 9586.27),
                (HoleType.OPEN_HOLE, 6.75, 5.12, 9586.27, 9618.66),
                (HoleType.OPEN_HOLE, 6.75, 4.88, 9618.66, 9635.90),
                (HoleType.OPEN_HOLE, 6.75, 5.06, 9635.90, 9642.27),
                (HoleType.OPEN_HOLE, 6.75, 4.81, 9642.27, 9647.48),
                (HoleType.OPEN_HOLE, 6.75, 4.75, 9647.48, 9651.15),
                (HoleType.OPEN_HOLE, 6.75, 5.19, 9651.15, 9680.84),
                (HoleType.OPEN_HOLE, 6.75, 5.25, 9680.84, 9685.21),
                (HoleType.OPEN_HOLE, 6.75, 4.81, 9685.21, 9700.0),
            )
        }

        wellbore = run_drillstring_and_create_wellbore(wits)
        hole = wellbore.actual_hole
        annulus = wellbore.actual_annulus

        # Assert hole configuration
        print("\n Hole configuration is:")
        print(hole)
        self.assert_hole(expected_results.get("hole"), hole)

        # Assert annulus configuration
        print("\n Annulus configuration is:")
        print(annulus)
        self.assert_annulus(expected_results.get("annulus"), annulus)

    def test_detailed_liner_while_running(self):
        os.environ["LOGGING_LEVEL"] = "debug"
        os.environ["LOGGING_ASSET_ID"] = "all"

        asset_id = 13684306
        Logger.set_asset_id(asset_id)

        wits = {
            "asset_id": asset_id,
            "metadata": {
                "casing": "5fbe0ec8eb6ace1f53efdcfe"
            },
            "data": {
                "bit_depth": 8_000.0,
                "hole_depth": 8_568.0
            }
        }

        expected_results = {
            "hole": (
                (HoleType.CASED_HOLE, 8.84, 0.0, 5710.0),
                (HoleType.OPEN_HOLE, 8.75, 5710.0, 8568.0)
            ),
            "annulus": (
                (HoleType.CASED_HOLE, 8.84, 5.0, 0, 5017.41),
                (HoleType.CASED_HOLE, 8.84, 7.63, 5017.41, 5710.0),
                (HoleType.OPEN_HOLE, 8.75, 7.63, 5710.0, 7956.0),
                (HoleType.OPEN_HOLE, 8.75, 7.5, 7956.0, 8000.0)
            )
        }

        wellbore = run_casingstring_and_create_wellbore(wits)
        hole = wellbore.actual_hole
        annulus = wellbore.actual_annulus

        # Assert hole configuration
        print("\n Hole configuration is:")
        print(hole)
        self.assert_hole(expected_results.get("hole"), hole)

        # Assert annulus configuration
        print("\n Annulus configuration is:")
        print(annulus)
        self.assert_annulus(expected_results.get("annulus"), annulus)

    def test_riser_boostline(self):
        os.environ["LOGGING_LEVEL"] = "debug"
        os.environ["LOGGING_ASSET_ID"] = "all"

        asset_id = 35629349
        Logger.set_asset_id(asset_id)

        wits = {
            "asset_id": asset_id,
            "metadata": {
                "drillstring": "5f8f46edb2ce18420cc43eb4"
            },
            "data": {
                "bit_depth": 21_000.0,
                "hole_depth": 21_000.0
            }
        }

        expected_results = {
            "hole": (
                (HoleType.CASED_HOLE, 7.0, 0.0, 10142.0),
                (HoleType.CASED_HOLE, 6.5, 10142.0, 20551.0),
                (HoleType.CASED_HOLE, 6.5, 20551.0, 20591.0),
                (HoleType.OPEN_HOLE, 6.0, 20591.0, 21000.0)
            ),
            "annulus": (
                (HoleType.CASED_HOLE, 7.0, 4.0, 0, 10142.0),
                (HoleType.CASED_HOLE, 6.5, 4.0, 10142.0, 17858.88),
                (HoleType.CASED_HOLE, 6.5, 4.75, 17858.88, 17870.02),
                (HoleType.CASED_HOLE, 6.5, 5.0, 17870.02, 17881.59),
                (HoleType.CASED_HOLE, 6.5, 4.0, 17881.59, 20591.0),
                (HoleType.OPEN_HOLE, 6.0, 4.0, 20591.0, 20778.74),
                (HoleType.OPEN_HOLE, 6.0, 4.9, 20778.74, 20783.23),
                (HoleType.OPEN_HOLE, 6.0, 4.0, 20783.23, 20846.14),
                (HoleType.OPEN_HOLE, 6.0, 4.81, 20846.14, 20849.25),
                (HoleType.OPEN_HOLE, 6.0, 4.75, 20849.25, 20852.34),
                (HoleType.OPEN_HOLE, 6.0, 5.0, 20852.34, 20883.62),
                (HoleType.OPEN_HOLE, 6.0, 4.75, 20883.62, 20913.60),
                (HoleType.OPEN_HOLE, 6.0, 4.56, 20913.60, 20941.39),
                (HoleType.OPEN_HOLE, 6.0, 4.62, 20941.39, 20969.18),
                (HoleType.OPEN_HOLE, 6.0, 5.0, 20969.18, 21000.0),
            )
        }

        wellbore = run_drillstring_and_create_wellbore(wits)
        hole = wellbore.actual_hole
        annulus = wellbore.actual_annulus

        # Assert hole configuration
        print("\n Hole configuration is:")
        print(hole)
        self.assert_hole(expected_results.get("hole"), hole)

        # Assert annulus configuration
        print("\n Annulus configuration is:")
        print(annulus)
        self.assert_annulus(expected_results.get("annulus"), annulus)

    def test_riserless(self):
        wits_json_file = os.path.abspath(
            os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources", "test", "wits_riserless.json")
        )
        wits = file_to_json(wits_json_file)
        wellbore = run_drillstring_and_create_wellbore(wits)

        hole: Hole = wellbore.actual_hole

        self.assertEqual(3, len(hole))

        self.assertEqual(HoleType.CASED_HOLE, hole[0].hole_type)
        self.assertAlmostEqual(3000.0, hole[0].top_depth, 1)
        self.assertAlmostEqual(10000.0, hole[0].bottom_depth, 1)
        self.assertAlmostEqual(7000.0, hole[0].length, 1)
        self.assertAlmostEqual(12.415, hole[0].inner_diameter, delta=0.01)

        # Under reamer open hole
        self.assertEqual(HoleType.OPEN_HOLE, hole[1].hole_type)
        self.assertAlmostEqual(10_000.0, hole[1].top_depth, 1)
        self.assertAlmostEqual(14_906.77, hole[1].bottom_depth, delta=0.1)
        self.assertAlmostEqual(4_906.77, hole[1].length, delta=0.1)
        self.assertAlmostEqual(12.0, hole[1].inner_diameter, delta=0.01)

        # Bit sized open hole
        self.assertEqual(HoleType.OPEN_HOLE, hole[2].hole_type)
        self.assertAlmostEqual(14_906.77, hole[2].top_depth, delta=0.1)
        self.assertAlmostEqual(15000, hole[2].bottom_depth, 1)
        self.assertAlmostEqual(93.23, hole[2].length, delta=0.1)
        self.assertAlmostEqual(8.5, hole[2].inner_diameter, delta=0.01)

        drillstring = wellbore.actual_drillstring
        self.assertAlmostEqual(0.0, drillstring[0].top_depth, 1)
        self.assertAlmostEqual(15_000.0, drillstring[-1].bottom_depth, 1)

        annulus = wellbore.actual_annulus
        self.assertEqual(HoleType.CASED_HOLE, annulus[0].hole_type)
        self.assertAlmostEqual(3_000.0, annulus[0].top_depth, 1)
        self.assertAlmostEqual(10_000.0, annulus[0].bottom_depth, 1)
        self.assertAlmostEqual(12.415, annulus[0].inner_diameter_hole, delta=0.01)

        self.assertEqual(HoleType.OPEN_HOLE, annulus[1].hole_type)
        self.assertAlmostEqual(10_000.0, annulus[1].top_depth, 1)
        self.assertAlmostEqual(14_895.9, annulus[1].bottom_depth, 1)
        self.assertAlmostEqual(12.0, annulus[1].inner_diameter_hole, delta=0.01)

        self.assertEqual(HoleType.OPEN_HOLE, annulus[3].hole_type)
        self.assertAlmostEqual(14906.77, annulus[3].top_depth, 1)
        self.assertAlmostEqual(15_000.0, annulus[3].bottom_depth, 1)
        self.assertAlmostEqual(8.5, annulus[3].inner_diameter_hole, delta=0.01)

    def assert_annulus(self, expected_annulus: tuple, actual_annulus: Annulus):
        expected_number_of_sections = len(expected_annulus)
        self.assertEqual(expected_number_of_sections, len(actual_annulus))

        for i in range(expected_number_of_sections):
            actual_annulus_section = (
                actual_annulus[i].hole_type,
                actual_annulus[i].inner_diameter_hole,
                round(actual_annulus[i].outer_diameter_pipe, 2),
                round(actual_annulus[i].top_depth, 2),
                round(actual_annulus[i].bottom_depth, 2)
            )
            expected_annulus_section = expected_annulus[i]
            self.assertEqual(expected_annulus_section, actual_annulus_section)

    def assert_hole(self, expected_hole: tuple, actual_hole: Hole):
        expected_number_of_sections = len(expected_hole)
        self.assertEqual(expected_number_of_sections, len(actual_hole))

        for i in range(expected_number_of_sections):
            actual_hole_section = (
                actual_hole[i].hole_type,
                actual_hole[i].inner_diameter,
                round(actual_hole[i].top_depth, 2),
                round(actual_hole[i].bottom_depth, 2)
            )
            expected_hole_section = expected_hole[i]
            self.assertEqual(expected_hole_section, actual_hole_section)

    def run_under_reamer_case(self, wits, expected_values):
        wellbore = run_drillstring_and_create_wellbore(wits)

        # Activate under-reamer in cased hole
        # This should raise an alert in console with WARN
        bit_depth = 9000.17
        hole_depth = 15005.17
        mud_flow_rate = 521.25
        wellbore.update(bit_depth, hole_depth, mud_flow_rate)

        # Update to activation depth and increase mud flow rate
        bit_depth = 15005.17
        hole_depth = 15005.17
        mud_flow_rate = 521.25
        wellbore.update(bit_depth, hole_depth, mud_flow_rate)

        # Update to target depth
        bit_depth = 20633.00
        hole_depth = 20633.00
        mud_flow_rate = 521.25
        wellbore.update(bit_depth, hole_depth, mud_flow_rate)

        # test serialization
        ser_str = obj2json(wellbore)
        wellbore_des = json2obj(ser_str)
        self.assertEqual(len(wellbore.actual_drillstring), len(wellbore_des.actual_drillstring))
        self.assertAlmostEqual(wellbore.actual_drillstring[1].inner_diameter, wellbore_des.actual_drillstring[1].inner_diameter, 10)
        self.assertAlmostEqual(wellbore.actual_drillstring[1].top_depth, wellbore_des.actual_drillstring[1].top_depth, 10)

        hole = wellbore.actual_hole
        annulus = wellbore.actual_annulus

        # Assert hole configuration
        print("\n Hole configuration is:")
        print(hole)

        self.assert_hole(expected_values.get("hole"), hole)

        # Assert annulus configuration
        print("\n Annulus configuration is:")
        print(annulus)
        self.assert_annulus(expected_values.get("annulus"), annulus)


def get_wits():
    wits_json_file = os.path.abspath(
        os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources", "test", "wits.json")
    )
    return file_to_json(wits_json_file)


def reset_drillstring(activation_type="flow_activated", activation_depth=None):
    if activation_type == "unactivated":
        activation_type = "flow_activated"
        activation_depth = 10_000_000.0

    elif activation_type == "depth_activation":
        activation_depth = 15205.17

    api_worker = API()
    path = "/v1/data/corva/data.drillstring/5ebaceeac93924003a922593"
    drillstring = api_worker.get(path=path).data
    drillstring["data"]["start_depth"] = 9891
    for component in drillstring["data"]["components"]:
        if component.get("family") == "ur":
            component["ur_opened_depth"] = activation_depth
            component["activation_logic"] = activation_type
            break

    api_worker.post(path="/v1/data/corva/data.drillstring", data=json.dumps(drillstring))



import copy
import os
import unittest
from worker.event.event_handler import EventHandler
from worker.test.utils import file_to_json

from worker.framework import constants


default_parameters = copy.deepcopy(constants.parameters)


class TestEvents(unittest.TestCase):
    def _load_event(self, filename: str) -> dict:
        file_path = os.path.abspath(
            os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources", "test", "events", filename)
        )
        event = file_to_json(file_path)
        return event

    def test_stream_events(self):
        constants.update({
            "global": {
                "event-type": "wits_stream",
                "post-to-message-producer": True,
            }
        })

        wits_stream = self._load_event("wits_stream.json")

        event_handler = EventHandler(app=None)
        event_handler._load(wits_stream)

        expected_output = {
            # asset_id: length of events
            265: 22,
            1000: 6
        }

        self.assertTrue(event_handler.event_by_asset_id[265].is_posting_to_message_producer)

        for asset_id, stream in event_handler.event_by_asset_id.items():
            self.assertEqual(expected_output[asset_id], len(stream))

        constants.parameters = copy.deepcopy(default_parameters)

    def test_scheduled_events(self):
        constants.update({"global": {"event-type": "scheduler"}})

        scheduler = self._load_event("scheduler.json")

        event_handler = EventHandler(app=None)
        event_handler._load(scheduler)

        expected_output = {
            # asset_id: length of events
            1: 9,
            2: 2,
            3: 1
        }

        for asset_id, records in event_handler.event_by_asset_id.items():
            self.assertEqual(expected_output[asset_id], len(records))

        constants.parameters = copy.deepcopy(default_parameters)



from typing import Union, List

from worker.event import Event, EventType


class SingleScheduledEvent:
    def __init__(self, single_stream_event: dict):
        self.asset_id = single_stream_event.get('asset_id') or 0

        self.start_time = single_stream_event.get('schedule_start', 0) / 1000
        self.end_time = single_stream_event.get('schedule_end', 0) / 1000
        self.schedule_id = single_stream_event.get('schedule')

    def complete_event(self, api) -> None:
        """Sets schedule as completed."""
        if not self.schedule_id:
            return

        api.post(path=f'/scheduler/{self.schedule_id}/completed')


class ScheduledEvent(Event):
    """A scheduled event class that holds the events of a single asset_id."""
    def __init__(self, new_records: Union[SingleScheduledEvent, List[SingleScheduledEvent]]):
        super().__init__(EventType.SCHEDULER)

        if new_records:
            self.add_records(new_records)

    def add_records(self, new_records: Union[SingleScheduledEvent, List[SingleScheduledEvent]]):
        if isinstance(new_records, SingleScheduledEvent):
            new_records = [new_records]

        if not self.records:
            self.asset_id = new_records[0].asset_id

        super().add_records(new_records)

    def complete_event(self, api) -> None:
        for each in self.records:
            each.complete_event(api)



import json
from typing import List, Union

from cached_property import cached_property

from worker.data.operations import get_data_by_path
from worker.event import Event, EventType
from worker.framework import constants


class StreamEvent(Event):
    """A stream event class that holds the events of a single asset_id."""
    def __init__(self, event: dict, is_posting_to_message_producer: bool = False):
        super().__init__(EventType.WITS_STREAM)

        self.metadata = event.get('metadata') or {}
        self.records = event.get('records') or []

        self.asset_id = 0
        if self.records:
            self.asset_id = self.records[0].get('asset_id')

        app_key = constants.get('global.stream_app_key')
        self.app_connection_id = get_data_by_path(
            self.metadata,
            f'apps.{app_key}.app_connection_id',
            func=int,
            default=None
        )

    @cached_property
    def is_posting_to_message_producer(self) -> bool:
        """
        Whether to post to message producer for the wits stream or not.
        If there is no lambda app following your app then this should be false,
        because this process is intracting with the API which slows the process down.
        """
        is_posting: bool = constants.get('global.post-to-message-producer', False)
        return is_posting

    @classmethod
    def merge(cls, events: List['StreamEvent']) -> Union['StreamEvent', None]:
        """Merge stream events of the same asset id into one"""
        if not events:
            return None

        merged_event = events[0]
        for each in events[1:]:
            merged_event.add(each)

        return merged_event

    def add(self, other: 'StreamEvent'):
        if self.asset_id != other.asset_id:
            raise Exception(f"Events of different assets can not be merged; {self.asset_id} and {other.asset_id}!")

        self.add_records(other.records)

    def build_message_producer_payload(self) -> dict:
        return {
            'app_connection_id': self.app_connection_id,
            'asset_id': self.asset_id,
            'data': self.records
        }

    def complete_event(self, api):
        if not self.is_posting_to_message_producer or not self.app_connection_id:
            return

        payload = self.build_message_producer_payload()
        api.post(path='/v1/message_producer', data=json.dumps(payload))



# flake8: noqa: E402

"""
This test is created to make sure the framework for local testing
is working correctly.
"""

# added the main directory as the first path
from os.path import dirname
import sys

parent = dirname(dirname(__file__))
sys.path.insert(0, parent)

from testcase.app.disable_module import DisableModule
from testcase.app.drilling_efficiency_app import DrillingEfficiency
from testcase.app import app_lambda
from worker.test import AppTestRun


if __name__ == '__main__':

    # turning off the "AnotherModule" for the testing purpose
    DisableModule.enabled = False

    drilling_efficiency_app = DrillingEfficiency()
    active_modules = drilling_efficiency_app.get_active_modules()

    collections = [module.collection for module in active_modules]

    # loading of constants should happen here to avoid conflicts
    from testcase.app.app_constants import constants

    app = AppTestRun(app_lambda.lambda_handler, collections, constants)
    app.run()



import json
import os
import unittest

from worker.data.activity import Activity
from worker.data.enums import DataStatus, ChannelStatus
from worker.data.operations import compute_time_step
from worker.data.wits import serialize, Channel, WITS
from worker.test.utils import file_to_json


class TestWits(unittest.TestCase):
    def test_set_wits(self):
        wits_sample = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'resources', 'test', 'wits_class',
                                                   'wits_sample1.json'))
        wits_json_list = file_to_json(wits_sample)

        wits = WITS.set_wits(wits_json_list)
        self.assertIsInstance(wits, list)

        wits_json_record = wits_json_list[0]
        wits = WITS.set_wits(wits_json_record)
        self.assertIsInstance(wits, WITS)

        self.assertEqual(1588280468, wits.timestamp)
        self.assertEqual(1588280468, wits.entry_at)
        self.assertEqual(12989, wits.hole_depth)
        self.assertEqual(12981.8, wits.bit_depth)
        self.assertEqual(113.2, wits.block_height)
        self.assertEqual(190.8, wits.hook_load)
        self.assertEqual(31.5, wits.weight_on_bit)
        self.assertEqual(0, wits.rop)
        self.assertEqual(0.14, wits.rotary_rpm)
        self.assertEqual(0.01898, wits.rotary_torque)
        self.assertEqual(596.13, wits.mud_flow_in)
        self.assertEqual(22.3, wits.mud_flow_out_percent)
        self.assertEqual(5148.15, wits.standpipe_pressure)
        self.assertEqual(93.37, wits.diff_press)
        self.assertEqual(96.05, wits.pump_spm_1)
        self.assertEqual(96.28, wits.pump_spm_2)
        self.assertIsNone(wits.pump_spm_3)
        self.assertEqual(192.32999999999998, wits.pump_spm_total)
        self.assertIsNone(wits.active_pit_volume)
        self.assertEqual(Activity.WASHING_DOWN, wits.state)

        self.assertEqual(DataStatus.VALID, wits.hole_depth.status)

        setattr(wits, "pump_spm_3", Channel(90.0, DataStatus.MISSING))
        self.assertEqual(90.0, wits.pump_spm_3)
        self.assertEqual(DataStatus.MISSING, wits.pump_spm_3.status)

        # converting wits to json
        wits_dict = wits.get_as_dict()
        self.assertEqual(1588280468, wits_dict['timestamp'])
        wits_dict_data = wits_dict['data']
        self.assertEqual(1588280468, wits_dict_data['entry_at'])
        self.assertEqual(5148.15, wits_dict_data['standpipe_pressure'])
        self.assertEqual("Washing Down", wits_dict_data['state'])

        # Test Deep Copy
        copied_wits = WITS.set_wits(serialize(wits))
        self.assertTrue(wits.mud_flow_out_percent == copied_wits.mud_flow_out_percent == 22.3)

    def test_wits_range_validation(self):
        wits_sample = os.path.abspath(
            os.path.join(os.path.dirname(__file__), '..', 'resources', 'test', 'wits_class', 'wits_sample1.json')
        )
        wits_json_list = file_to_json(wits_sample)

        wits = WITS.set_wits(wits_json_list)
        self.assertIsInstance(wits, list)

        wits_json_record = wits_json_list[0]
        wits = WITS.set_wits(wits_json_record)
        self.assertIsInstance(wits, WITS)

        self.assertEqual(1588280468, wits.timestamp)
        # data qc and clipping
        wits.mud_flow_out_percent = -6
        channel_range_mapping = {
            'mud_flow_out_percent': {
                'range': [0, 100],
                'operation': 'clip'
            }
        }
        wits.validate_range(channel_range_mapping)
        self.assertEqual(0, wits.mud_flow_out_percent)

    def test_get_mud_flow_in_status(self):
        wits_sample = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'resources', 'test', 'wits_class',
                                                   'wits_sample1.json'))
        wits_sample = file_to_json(wits_sample)[0]

        wits = WITS.set_wits(wits_sample)

        self.assertEqual(ChannelStatus.ON, wits.mud_flow_in_status(low_mud_flow_in_threshold=20))
        self.assertEqual(ChannelStatus.OFF, wits.mud_flow_in_status(low_mud_flow_in_threshold=800))

        wits.mud_flow_in = None
        self.assertEqual(ChannelStatus.MISSING, wits.mud_flow_in_status(low_mud_flow_in_threshold=800))

    def test_get_spm_status(self):
        wits_sample = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'resources', 'test', 'wits_class',
                                                   'wits_sample1.json'))
        wits_sample = file_to_json(wits_sample)[0]
        wits = WITS.set_wits(wits_sample)

        self.assertEqual(ChannelStatus.ON, wits.spm_status(low_spm_threshold=20))
        self.assertEqual(ChannelStatus.OFF, wits.spm_status(low_spm_threshold=200))

        wits.pump_spm_1 = 0
        self.assertEqual(ChannelStatus.ON, wits.spm_status(low_spm_threshold=20))

        wits.pump_spm_1 = None
        wits.pump_spm_2 = None
        self.assertEqual(ChannelStatus.MISSING, wits.spm_status(low_spm_threshold=100))

        wits_sample = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'resources', 'test', 'wits_class',
                                                   'wits_sample2.json'))
        wits_sample = file_to_json(wits_sample)[0]
        wits = WITS.set_wits(wits_sample)

        self.assertEqual(ChannelStatus.OFF, wits.spm_status(low_spm_threshold=20))

    def test_compute_time_step(self):
        """
        Computing the time step with two method:
        - a list of wits records in WITS object format
        - a list of wits records in json format
        :return:
        """
        wits_chunk = os.path.abspath(
            os.path.join(os.path.dirname(__file__), '..', 'resources', 'test', 'activity_grouping.data')
        )

        with open(wits_chunk, 'r') as file:
            lines = file.readlines()
            records_dict = [json.loads(record) for record in lines]
            records_wits = [WITS(json.loads(record)) for record in lines]

            timestep_dict = compute_time_step(records_dict)
            timestep_wits = compute_time_step(records_wits)

            self.assertEqual(1, timestep_dict)
            self.assertEqual(1, timestep_wits)
